#!/bin/env python

from builtins import str
from builtins import map
import threading
import os
import glob
import errno
from time import time
import operator
from itertools import chain
import cs.Tools.Constants as Constants
from cs.Tools.utils import thread_id_string
from cs.Tools.LogConfig import enable_file_logging
from functools import reduce #pylint: disable=redefined-builtin

class DeleteThread(threading.Thread):


    def __init__(self,conf,event,dbLock,DeleteQueue,ClearQueue):

        threading.Thread.__init__(self, name="DeleteThread")

        self.conf = conf
        self.event = event
        self.dbLock = dbLock
        self.DeleteQueue = DeleteQueue
        self.ClearQueue = ClearQueue
        self.files_deleted = 0


        ##### Get configuration parameters #####
        self.lockFile = self.conf.lockFile
        self.ignoreLock = self.conf.ignoreLock
        self.nFiles = self.conf.nFiles
        self.minFileTime = self.conf.minFileTime
        self.migFlag = self.conf.migFlag
        self.highCriticMark = self.conf.highCriticMark
        self.lowCriticMark = self.conf.lowCriticMark
        self.lowWaterMark = self.conf.lowWaterMark
        self.DirList = [os.path.normpath(k) for k in self.conf.DirList]
        self.DeleteTimeout = self.conf.DeleteTimeout
        self.MigDelay = self.conf.MigDelay

        #### Import the function to resolve the merged file names #####
        self.mergedChecker = None
        if self.conf.MergedModule:
            self.mergedChecker = getattr(self.conf.MergedModule, \
                                         self.conf.MergedNameModule)

        self.MigDict = {}
        self.exitFlag = 0
        self.db = None
        self.dbFlag = True
        self.deleteNow = False

        # get logger to files, (Will also log to ERS)
        self.logger = enable_file_logging("DeleteThread","DeleteLog.out", conf.LogDir, conf.LogLevel)

    # end def __init__()

    def delete(self, deletefile):
        filename = deletefile[0]

        try:
            os.remove(filename)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                # The file has been removed by an external cause.
                # This is bad but this is not a reason to crash.
                self.logger.critical('error deleting %s: not present (deleted by external cause)', filename)
            else:
                raise exc

        try:
            os.remove(filename + Constants.copied_ext)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                # The .COPIED file has been removed by an external cause.
                # This is bad but this is not a reason to crash.
                self.logger.critical('error deleting %s: not present (deleted by external cause)', filename + Constants.copied_ext)
            else:
                raise exc

        ##### Update Oracle Metadata Database, if connection is fine #####
        ##### File table: filestate from CLOSED to DELETED #####
        if self.db:
            self.logger.info('Update Metadata database')
            self.dbLock.acquire()
            self.dbFlag = self.db.Deletion(filename)
            self.dbLock.release()
        elif self.conf.connection:
            self.logger.warning('No connection to Metadata database: database will not be updated for file %s', filename)
            # end if,else

        #increment files deleted, used by the Information Service Thread
        self.files_deleted = self.files_deleted + 1

        ##### Put .data deleted file in the clear queue #####
        self.ClearQueue.put(deletefile)
        #block until room is available in the queue ?????


    def run(self):

        self.logger.info(thread_id_string())

        fileCounter = 0
        fsInfo = self.getFSsize()

        while not self.exitFlag:

            if self.DeleteQueue.qsize() == 0:
                self.logger.debug('DeleteQueue is empty')
                self.event.wait(self.DeleteTimeout)
                continue
            # end if

            ##### Get file to be deleted from DeleteQueue #####
            ##### deletefile is a list of (.COPIED file, Castor pool, Castor copy directory, Castor stage host) #####
            deletefile = self.DeleteQueue.get(0)

            #deletefile = self.DeleteQueue.get()
            #block until room is available in the queue ?????
            self.logger.debug('Get file %s from DeleteQueue', deletefile[0])
            fileCounter += 1

            ##### Get Castor environment and copy directory #####
            filename = deletefile[0]
            pool = deletefile[1]
            copyDir = deletefile[2]
            stagehost = deletefile[3]

            ##### At beginning or after n copied files, get filesystem size info #####
            if fileCounter >= self.nFiles:
                self.logger.debug('After %d files it is time to get filesystem size info', fileCounter)
                fsInfo = self.getFSsize()
                fileCounter = 1
            # end if

            fsTot = fsInfo[0]
            fsocc = fsInfo[1]

            self.logger.debug('FS occupancies: %s', str(fsocc))
            self.logger.debug('Total occupancy: %f', fsTot)

            try:
                fsSingle = fsocc[DeleteThread.getmountpoint(filename)]
            except KeyError:
                ## DirList possibly changed in the meanwhile
                ## take the safest choice
                fsSingle = 0.

            self.logger.debug('Occupancy for %s filesystem: %f', filename, fsSingle)

            ##### Check if total filesystem size is over threshold #####
            if fsTot >= self.highCriticMark: self.deleteNow = True
            if fsTot <= self.lowCriticMark: self.deleteNow = False

            if self.deleteNow:
                self.logger.info('Critical deletion: delete file %s', filename)
                self.delete(deletefile)
                continue
            # end if

            ##### Do not delete file if it's too recent #####
            mtime = time() # conservative: if something goes wrong, file will be too recent to be deleted
            try:
                mtime = os.stat(filename)[8]
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    # The file has been removed by an external cause: log and finish the job
                    self.logger.critical('error accessing (stat) %s: not present (deleted by external cause): forcing complete deletion', filename)
                    self.delete(deletefile)
                    continue
                else:
                    raise exc

            if (time() - mtime) < self.minFileTime :
                self.DeleteQueue.put(deletefile)

                # block until room is available in the queue ?????
                self.event.wait(self.DeleteTimeout)
                self.logger.info('File %s too recent: delete it later', filename)
                continue

            directory = os.path.dirname(filename)

            ##### Check the SFO lock, if required #####
            if not self.ignoreLock:
                self.logger.debug('Check for locked filesystem required by the deletion policy')

                if glob.glob(os.path.join(directory, self.lockFile)):
                    ##### Do not delete: put file back into the DeleteQueue #####
                    self.DeleteQueue.put(deletefile)

                    # block until room is available in the queue ?????
                    self.logger.debug('Filesystem locked (it was unlocked for the Copy) : do not delete')
                    self.event.wait(self.DeleteTimeout)
                    continue
                # end if
            # end if not

            ##### Compare filesystem disk usage and Low WaterMark  #####

            ## Find the index of 'directory' in DirList. Be careful
            ## because DirList might contain wildcards
            try:
                idx = [idx for idx, dirlist_item in enumerate(self.DirList)
                       if directory in map(os.path.normpath,glob.iglob(dirlist_item))][0]
            except IndexError:
                ##DirList was possibly modified
                ##take safest choice
                watermark = 100.
            else:
                watermark = self.lowWaterMark[idx]

            if fsSingle < watermark and \
                   not watermark == 0:
                ##### Do not delete: put file back into the DeleteQueue #####
                self.DeleteQueue.put(deletefile)

                # block until room is available in the queue ?????
                self.logger.debug('Filesystem used size (%f %%) under threshold (%f %%): do not delete', fsSingle, watermark)
                self.event.wait(self.DeleteTimeout)
                continue
            # end if

            ##### Check if migration is required #####
            if not self.migFlag:
                self.logger.debug('Migration not required by the deletion policy: delete file %s', filename)

                self.delete(deletefile)

                continue
            # end if

            self.logger.debug('Migration required by the deletion policy')

            ##### Check if file has already been migrated only if it's time #####
            if not self.checkMigrationTimeout(filename):
                self.logger.debug('Last check for file to be migrated was too recent: do not delete')
                self.DeleteQueue.put(deletefile)

                self.event.wait(self.DeleteTimeout)
                continue
            # end if

            Castor_file = os.path.join(copyDir, os.path.basename(filename))

            (isMigr, isMerged) = \
                     self.checkMigration(Castor_file, stagehost, pool)

            if isMigr :
                self.logger.debug('File migrated: going to delete file %s', filename)

                ### For merged file the checksum check must be skipped
                ### since original files are not migrated

                success = True

                if not isMerged:

                    sizechecksum = self.conf.backend.sizechecksum
                    castorfilesize, checksum = \
                                    sizechecksum (Castor_file,\
                                                  stagehost, pool, \
                                                  self.logger)

                    sfofilesize = os.stat(filename)[6]

                    success = (sfofilesize == castorfilesize)

                    if self.db:

                        self.dbLock.acquire()
                        dbchecksum,dbfilesize,filehealth = self.db.FileInfo(filename)
                        self.dbLock.release()


                        if filehealth:
                            success &= (castorfilesize == dbfilesize)
                            success &= (dbchecksum.lower() == checksum.lower())
                        else:
                            success = False

                if success:
                    self.delete(deletefile)
                else:
                    self.logger.warning('Checksum check or size check failed for: %s'\
                                        '. Not deleting', filename)

                del self.MigDict[filename]
                #block until room is available in the queue ?????
                continue

            else:
                ##### File not migrated yet, put it back into the DeleteQueue (to retry later) #####
                self.logger.debug('File not migrated yet: do not delete and check later: %s', filename)
                self.DeleteQueue.put(deletefile)

            # end if,else

            self.event.wait(self.DeleteTimeout)

        # end while
        self.logger.info('DeleteThread exited')
    # end def run()


    def checkMigration(self, castorfile, stagehost, pool):
        ##### Check if file has already been migrated #####
        self.logger.debug('Time to check if file has been migrated: %s', castorfile)

        isMig = self.conf.backend.migrated(castorfile, stagehost,
                                           pool, self.logger)

        isMerged = False

        if (not isMig) and self.mergedChecker:
            self.logger.debug('Time to check if merged file has been migrated. Original file is %s',
                              castorfile)

            if self.mergedChecker(castorfile, stagehost, pool, \
                                  self.conf.backend, self.logger):
                isMig = True
                isMerged = True


        return isMig,isMerged


    ##### Get total size of all filesystems #####
    def getFSsize(self):

        ##### Normal case #####
        fs_dict = {}
        fs_occupancy = {}
        for dirlist_item in chain(*map(glob.iglob,self.DirList)):
            mount = DeleteThread.getmountpoint(dirlist_item)
            try:
                fs = os.statvfs(mount)
            except OSError as e:
                self.logger.error('Cannot stat %s: %s', e.filename, e.strerror)
                continue

            ## Use dictionary to avoid counting multiple times the same filesystem ##
            fs_dict[mount]=(float(fs.f_blocks),float(fs.f_bavail))
            fs_occupancy[mount] = 100.*(1-float(fs.f_bavail)/float(fs.f_blocks))

        totBlocks = reduce(operator.add, list(map(operator.itemgetter(0),
                                             list(fs_dict.values()))),
                           0)

        totFreeBlocks = reduce(operator.add, list(map(operator.itemgetter(1),
                                                 list(fs_dict.values()))),
                               0)

        ##### df: round to the next highest integer #####
        totFS = 0
        if totBlocks:
            totFS = 100.*(1-(float(totFreeBlocks)/float(totBlocks)))
        return (totFS, fs_occupancy)


    ##### Check if it's time to check if file is migrated #####
    ##### MigDict is a dictionary of {deletefile: timestamp} #####
    def checkMigrationTimeout(self,migFile):

        if migFile not in self.MigDict:
            self.MigDict[migFile] = time()
            return True
        # end if

        timestamp = self.MigDict.get(migFile)
        now = time()
        if (now-timestamp)>self.MigDelay :
            self.MigDict[migFile] = now
            return True
        else: return False
    # end def checkMig()



    ##### Exit handler #####
    def deleteExit(self):
        self.exitFlag = 1
        self.logger.info('Exit signal received')
    # end def deleteExit()


    ##### Set Oracle Metadata Database #####
    def setDB(self,db):
        self.db = db
        self.logger.debug('Set connection to MetaData database '
                         + 'as given by main thread')
    # end setDB()


    ##### Get Oracle Metadata Database #####
    def getDBFlag(self):
        self.logger.debug('Communicate to main thread if connection '
                          + 'to MetaData database exists')
        flag = self.dbFlag
        self.dbFlag = True
        return flag
    # end getDB()

    @staticmethod
    def getmountpoint(path):
        path = os.path.realpath(os.path.abspath(path))
        while path != os.path.sep:
            if os.path.ismount(path):
                return path
            path = os.path.abspath(os.path.join(path, os.pardir))
        return path
