#!/bin/env python

from __future__ import print_function
from builtins import object
import os.path
import cs.StorageBackends.castorstorage as castorstorage
import re
import importlib
from operator import itemgetter

def init(configuration):
    """
    In a configuration dictiornary is passed, we will look for the parameters for the castor backend:
    configuration["pool"] --> Castor pool
    configuration["stagehost"] --> Castor stager
    configuration["basedir"] --> base path in Castor
    configuration["nameparser"] --> nameparser module (optional)

    """

    nameparser = configuration.get('nameparser', 'SFOFileNameParser')

    module = importlib.import_module("cs.Tools.FilenameParsers." + nameparser)

    nameparser = getattr(module, nameparser)

    ddmmigrated = DDMLookup(configuration["pool"],
                            configuration["stagehost"],
                            configuration["basedir"],
                            nameparser)

    globals()['_ddmmigrated'] = ddmmigrated
    globals()['DDMMigrated'] = ddmmigrated._ddmMigrated #pylint: disable=protected-access


class DDMLookup(object):

    def __init__(self, pool, stagehost, basedir, parser):

        self.pool = pool
        self.stagehost = stagehost
        self.basedir = basedir
        self.parser = parser

    def _ddmMigrated(self, castorfile, stagehost, pool, backend,
                     logger=None, verbose=False):
        """
Destination directory:
/<basedir>/<projecttag>/<streamtype_streamname>/<runnr-8digit>/<dataset>/

basedir = /castor/cern.ch/grid/atlas/rucio/raw
dataset = data09_cos.00122050.physics_IDCosmic.merge.RAW

A file name is:
1) if all SFO files of a LB fit are merged into a single file
   <project>.<runnr>.<stream>.daq.RAW._lbXXXX._SFO-N._<fileseqnr>.data, N=4,5,9,10
   --> <project>.<runnr>.<stream>.merge.RAW._lbXXXX._SFO-ALL._0001.<attemptnr>

2) if all SFO files of a LB don't merge into a single file
   <project>.<runnr>.<stream>.daq.RAW._lbXXXX._SFO-N._<fileseqnr>.data, N=4,5,9,10
   --> <project>.<runnr>.<stream>.merge.RAW._lbXXXX._SFO-N._<fileseqnr>.<attemptnr>, N=4,5,9,10

3) if not merged
   <project>.<runnr>.<stream>.daq.RAW._lbXXXX._SFO-N._<fileseqnr>.data, N=4,5,9,10 --> <project>.<runnr>.<stream>.daq.RAW._lbXXXX._SFO-N._<fileseqnr>.data, N=4,5,9,10

All filenames may have a _<timestamp> appended in case of DDM transfer failure. In this case the large timestamp should be considered

debug_* are redirected to debug_all, but file names are preserved.

        """
        del stagehost, pool, backend # unused args

        original_name = os.path.basename(castorfile)

        ### Build the target directory name
        parsed = self.parser(original_name)

        streamname = parsed.StreamName()
        dataset = parsed.Dataset()

        if parsed.StreamType() == 'debug':
            streamname = 'all'
            dataset = dataset.replace("_%s" % parsed.StreamName(), \
                                      "_%s" % streamname)


        org_path = \
            os.path.join(self.basedir, parsed.ProjectTag(), \
                             '%s_%s' % (parsed.StreamType(),
                                        streamname), \
                             '%s' % parsed.RunNr(), \
                             '%s' % dataset)

        merged_path = \
           os.path.join(self.basedir, parsed.ProjectTag(), \
                            '%s_%s' % (parsed.StreamType(),
                                       streamname), \
                            '%s' % parsed.RunNr(), \
                            '%s' % dataset.replace(".daq.",".merge."))

        if verbose:
            print('Search path (ORG):',org_path)
            print('Search path (MER):',merged_path)


        ### Look for files into the dataset dir
        success, all_files = castorstorage.listdir(org_path, self.stagehost,
                                                   self.pool,
                                                   logger)

        if success:

            ### Match original filename + transfer errors
            regex = "^" + original_name + \
                "(?:_(?P<timestamp>[0-9]+))?$"
            regex = regex.replace('.','[.]')
            if verbose:
                print('Original file regex',regex)

            regex = re.compile(regex)
            match = regex.match

            found = [(int(m.groupdict(0)['timestamp']), m.group(0)) for f in all_files for m in (match(f),) if m]

            ### Take the most recent attempt (0 is the firt attempt)
            found.sort(key=itemgetter(0))
            if found:
                if verbose: print('Found:', found)
                return castorstorage.migrated(os.path.join(org_path,
                                                           found[-1][-1]),
                                              self.stagehost, self.pool, logger)

        ### Look for files into the dataset dir
        success, all_files = castorstorage.listdir(merged_path, self.stagehost,
                                                   self.pool,
                                                   logger)

        ### If fails, return false
        if not success: return False

        #^name[.](?:(?:_SFO-1._ZZZZ)|(?:_SFO-ALL._0001))[.](?P<t0attempt>[0-9]+)(?:_(?P<timestamp>[0-9]+))?$

        ### Create the two type of name we expect (neglecting attempt number)
        basename = os.path.splitext(original_name)[0]
        merged = basename.replace('.daq.','.merge.')

        original_sfo = merged.split('.',6)[-1]
        basename = '.'.join(merged.split('.',6)[:-1])

        regex = "^" + basename + \
            ".(?:(?:%s)|(?:_SFO-ALL._0001))."  % original_sfo + \
            "(?P<t0attempt>[0-9]+)(?:_(?P<timestamp>[0-9]+))?$"
        regex = regex.replace('.','[.]')
        if verbose:
            print('Merged file regex',regex)
        regex = re.compile(regex)
        match = regex.match

        ### The 't0attempt' must be always present hence
        ### we do not need a default
        found = [(int(m.groupdict()['t0attempt']), int(m.groupdict(0)['timestamp']), m.group(0)) for f in all_files for m in (match(f),) if m]

        found.sort(key=itemgetter(0,1))
        if found:
            if verbose: print('Found:', found)
            return castorstorage.migrated(os.path.join(merged_path,
                                                       found[-1][-1]),
                                          self.stagehost, self.pool, logger)
        return False


if __name__ == '__main__':

    def _parseargs():
        import argparse

        parser =\
            argparse.ArgumentParser(description='Check file tape migration',
                                    formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        parser.add_argument('-p', '--pool', default='t0atlas', help=' ')
        parser.add_argument('-s', '--stagehost', default='castoratlas',
	help=' ')
        parser.add_argument('-b', '--basedir',
                            default='/castor/cern.ch/grid/atlas/rucio/raw',
	help=' ')
        parser.add_argument('-n', '--nameparser', default=None)
        parser.add_argument('-f', '--file', required=True)

        return parser.parse_args()

    args = _parseargs()

    config = {'pool':args.pool, 'stagehost':args.stagehost,
              'basedir':args.basedir}

    if args.nameparser != None: config['nameparser'] = args.nameparser

    init(config)

    # pylint cannot find dynamically-added global function
    print(DDMMigrated(args.file, None, None, #pylint: disable=E0602
                      None, verbose=True))
